@@ -18,25 +18,32 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 #
+from __future__ import annotations
 
-from __future__ import print_function
+import errno
+import fcntl
+import itertools
+import os
+import pickle
+import re
+import shutil
+import socket
+import sys
+import textwrap
+import time
+from importlib.machinery import SourceFileLoader
+from operator import itemgetter
+from types import TracebackType
+from typing import TYPE_CHECKING, Any, Iterable, Iterator
+from urllib.parse import unquote
+
+import xapian
 
 import axi
-import sys
-import os
-import imp
-import socket, errno
-import fcntl
-import textwrap
-import xapian
-import shutil
-import itertools
-import time
-import re
-import pickle
 
-from operator import itemgetter
-from six.moves.urllib_parse import unquote
+if TYPE_CHECKING:
+    import apt
+    import apt_pkg
 
 APTLISTDIR="/var/lib/apt/lists"
 
@@ -44,13 +51,18 @@ class Addon:
     """
     Indexer plugin wrapper
     """
-    def __init__(self, fname, progress=None, **kw):
+    def __init__(
+        self,
+        fname: str,
+        progress=None,
+        **kw
+    ) -> None:
         self.filename = os.path.basename(fname)
         self.name = os.path.splitext(self.filename)[0]
         oldpath = sys.path
         try:
             sys.path.append(os.path.dirname(fname))
-            self.module = imp.load_source("axi.plugin_" + self.name, fname)
+            self.module = SourceFileLoader("axi.plugin_" + self.name, fname).load_module()
         finally:
             sys.path = oldpath
         try:
@@ -68,17 +80,17 @@ class Addon:
                     progress.warning("Plugin %s initialisation failed: %s" % (fname, str(e)))
                 self.obj = None
 
-    def finished(self):
+    def finished(self) -> None:
         if hasattr(self.obj, "finished"):
             self.obj.finished()
 
-    def send_extra_info(self, **kw):
+    def send_extra_info(self, **kw: Any) -> None:
         func = getattr(self.obj, "send_extra_info", None)
         if func is not None:
             func(**kw)
 
 class Plugins(list):
-    def __init__(self, **kw):
+    def __init__(self, **kw) -> None:
         """
         Read the plugins, in sorted order.
 
@@ -88,20 +100,23 @@ class Plugins(list):
             kw["langs"] = self.scan_available_languages()
         progress = kw.get("progress", None)
 
-        self.disabled = []
+        self.disabled: list[str] = []
         for fname in sorted(os.listdir(axi.PLUGINDIR)):
             # Skip non-pythons, hidden files and python sources starting with '_'
-            if fname[0] in ['.', '_'] or not fname.endswith(".py"): continue
+            if fname[0] in ['.', '_'] or not fname.endswith(".py"):
+                continue
             fullname = os.path.join(axi.PLUGINDIR, fname)
-            if not os.path.isfile(fullname): continue
-            if progress: progress.verbose("Reading plugin %s." % fullname)
+            if not os.path.isfile(fullname):
+                continue
+            if progress:
+                progress.verbose("Reading plugin %s." % fullname)
             addon = Addon(fullname, **kw)
             if addon.obj != None:
                 self.append(addon)
 
-    def scan_available_languages(self):
+    def scan_available_languages(self) -> set[str]:
         # Languages we index
-        langs = set()
+        langs: set[str] = set()
 
         # Look for files like: ftp.uk.debian.org_debian_dists_sid_main_i18n_Translation-it
         # And extract the language code at the end
@@ -118,104 +133,126 @@ class Progress:
     """
     Normal progress report to stdout
     """
-    def __init__(self):
-        self.task = None
+    def __init__(self) -> None:
+        self.task: str | None = None
         self.halfway = False
         self.is_verbose = False
-    def begin(self, task):
+
+    def begin(self, task: str) -> None:
         self.task = task
         print("%s..." % self.task, end='')
         sys.stdout.flush()
         self.halfway = True
-    def progress(self, percent):
+
+    def progress(self, percent: float) -> None:
         print("\r%s... %d%%" % (self.task, percent), end='')
         sys.stdout.flush()
         self.halfway = True
-    def end(self):
+
+    def end(self) -> None:
         print("\r%s: done.  " % self.task)
         self.halfway = False
-    def verbose(self, *args):
+
+    def verbose(self, *args: str) -> None:
         if not self.is_verbose: return
         if self.halfway:
             print()
         print(" ".join(args))
         self.halfway = False
-    def notice(self, *args):
+
+    def notice(self, *args: str) -> None:
         if self.halfway:
             print()
         print(" ".join(args), file=sys.stderr)
         self.halfway = False
-    def warning(self, *args):
+
+    def warning(self, *args: str) -> None:
         if self.halfway:
             print()
         print(" ".join(args), file=sys.stderr)
         self.halfway = False
-    def error(self, *args):
+
+    def error(self, *args: str) -> None:
         if self.halfway:
             print()
         print(" ".join(args), file=sys.stderr)
         self.halfway = False
 
+
 class BatchProgress:
     """
     Machine readable progress report
     """
-    def __init__(self):
-        self.task = None
-    def begin(self, task):
+    def __init__(self) -> None:
+        self.task: str | None = None
+
+    def begin(self, task: str) -> None:
         self.task = task
         print("begin: %s\n" % self.task, end='')
         sys.stdout.flush()
-    def progress(self, percent):
+
+    def progress(self, percent: float) -> None:
         print("progress: %d/100\n" % percent, end='')
         sys.stdout.flush()
-    def end(self):
+
+    def end(self) -> None:
         print("done: %s\n" % self.task)
         sys.stdout.flush()
-    def verbose(self, *args):
+
+    def verbose(self, *args: str) -> None:
         print("verbose: %s" % (" ".join(args)))
         sys.stdout.flush()
-    def notice(self, *args):
+
+    def notice(self, *args: str) -> None:
         print("notice: %s" % (" ".join(args)))
         sys.stdout.flush()
-    def warning(self, *args):
+
+    def warning(self, *args: str) -> None:
         print("warning: %s" % (" ".join(args)))
         sys.stdout.flush()
-    def error(self, *args):
+
+    def error(self, *args:str) -> None:
         print("error: %s" % (" ".join(args)))
         sys.stdout.flush()
 
+
 class SilentProgress:
     """
     Quiet progress report
     """
-    def begin(self, task):
+    def begin(self, task: str) -> None:
         pass
-    def progress(self, percent):
+
+    def progress(self, percent: float) -> None:
         pass
-    def end(self):
+
+    def end(self) -> None:
         pass
-    def verbose(self, *args):
+
+    def verbose(self, *args: str) -> None:
         pass
-    def notice(self, *args):
+
+    def notice(self, *args: str) -> None:
         pass
-    def warning(self, *args):
+    def warning(self, *args: str) -> None:
         print(" ".join(args), file=sys.stderr)
-    def error(self, *args):
+
+    def error(self, *args: str) -> None:
         print(" ".join(args), file=sys.stderr)
 
+
 class ClientProgress:
     """
     Client-side progress report, reporting progress from another running
     indexer
     """
-    def __init__(self, progress):
+    def __init__(self, progress) -> None:
         self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
         self.sock.settimeout(None)
         self.sock.connect(axi.XAPIANDBUPDATESOCK)
         self.progress = progress
 
-    def loop(self):
+    def loop(self) -> None:
         hasBegun = False
         while True:
             msg = self.sock.recv(4096)
@@ -255,38 +292,48 @@ class ServerSenderProgress:
     """
     Server endpoint for client-server progress report
     """
-    def __init__(self, sock, task = None):
+    def __init__(self, sock: socket.socket, task: str | None = None) -> None:
         self.sock = sock
-        self.task = task
-    def __del__(self):
+        self.task: str | None = task
+
+    def __del__(self) -> None:
         self._send(pickle.dumps(("alldone",)))
-    def _send(self, text):
+
+    def _send(self, text: bytes) -> None:
         try:
             self.sock.send(text)
         except:
             pass
-    def begin(self, task):
+
+    def begin(self, task: str) -> None:
         self.task = task
         self._send(pickle.dumps(("begin", self.task)))
-    def progress(self, percent):
+
+    def progress(self, percent: float) -> None:
         self._send(pickle.dumps(("progress", self.task, percent)))
-    def end(self):
+
+    def end(self) -> None:
         self._send(pickle.dumps(("end", self.task)))
-    def verbose(self, *args):
+
+    def verbose(self, *args: str) -> None:
         self._send(pickle.dumps(("verbose",) + args))
-    def notice(self, *args):
+
+    def notice(self, *args: str) -> None:
         self._send(pickle.dumps(("notice",) + args))
-    def warning(self, *args):
+
+    def warning(self, *args: str) -> None:
         self._send(pickle.dumps(("warning",) + args))
-    def error(self, *args):
+
+    def error(self, *args: str) -> None:
         self._send(pickle.dumps(("error",) + args))
 
+
 class ServerProgress:
     """
     Send progress report to any progress object, as well as to client indexers
     """
-    def __init__(self, mine):
-        self.task = None
+    def __init__(self, mine) -> None:
+        self.task: str | None = None
         self.proxied = [mine]
         self.sockfile = axi.XAPIANDBUPDATESOCK
         try:
@@ -301,10 +348,12 @@ class ServerProgress:
         # Disallowing unwanted people to mess with the file is automatic, as
         # the socket has the ownership of the user we're using, and people
         # can't connect to it unless they can write to it
-    def __del__(self):
+
+    def __del__(self) -> None:
         self.sock.close()
         os.unlink(self.sockfile)
-    def _check(self):
+
+    def _check(self) -> None:
         try:
             sock = self.sock.accept()[0]
             self.proxied.append(ServerSenderProgress(sock, self.task))
@@ -312,66 +361,81 @@ class ServerProgress:
             if e.args[0] != errno.EAGAIN:
                 raise
         pass
-    def begin(self, task):
+
+    def begin(self, task: str) -> None:
         self._check()
         self.task = task
         for x in self.proxied: x.begin(task)
-    def progress(self, percent):
+
+    def progress(self, percent: float) -> None:
         self._check()
         for x in self.proxied: x.progress(percent)
-    def end(self):
+
+    def end(self) -> None:
         self._check()
         for x in self.proxied: x.end()
-    def verbose(self, *args):
+
+    def verbose(self, *args: str) -> None:
         self._check()
         for x in self.proxied: x.verbose(*args)
-    def notice(self, *args):
+
+    def notice(self, *args: str) -> None:
         self._check()
         for x in self.proxied: x.notice(*args)
-    def warning(self, *args):
+
+    def warning(self, *args: str) -> None:
         self._check()
         for x in self.proxied: x.warning(*args)
-    def error(self, *args):
+
+    def error(self, *args: str) -> None:
         self._check()
         for x in self.proxied: x.error(*args)
 
 
-class ExecutionTime(object):
+class ExecutionTime:
     """
     Helper that can be used in with statements to have a simple
     measure of the timing of a particular block of code, e.g.
     with ExecutionTime("db flush"):
         db.flush()
     """
-    def __init__(self, info=""):
+    def __init__(self, info: str = "") -> None:
         self.info = info
-    def __enter__(self):
+
+    def __enter__(self) -> None:
         self.now = time.time()
-    def __exit__(self, type, value, stack):
+
+    def __exit__(
+        self,
+        type : type[BaseException] | None = None,
+        value: BaseException | None = None,
+        stack: TracebackType | None = None
+    ) -> None:
         print("%s: %s" % (self.info, time.time() - self.now))
 
-class Indexer(object):
+
+class Indexer:
     """
     The indexer
     """
-    def __init__(self, progress, quietapt=False):
+    def __init__(self, progress, quietapt: bool = False) -> None:
         self.progress = progress
         self.quietapt = quietapt
         self.verbose = getattr(progress, "is_verbose", False)
         # Timestamp of the most recent data source
         self.ds_timestamp = 0
         # Apt cache instantiated on demand
-        self.apt_cache = None
+        self.apt_cache: apt.Cache | None = None
         # Loaded plugins
-        self.plugins = None
+        self.plugins: list[Addon] | None = None
         # OS file handle for the lock file
-        self.lockfd = None
+        self.lockfd: int | None = None
 
         # Ensure the database and cache directories exist
         self.ensure_dir_exists(axi.XAPIANDBPATH)
         self.ensure_dir_exists(axi.XAPIANCACHEPATH)
 
-    def ensure_dir_exists(self, pathname):
+    def ensure_dir_exists(self, pathname: str) -> None:
         """
         Create a directory if missing, but do not complain if it already exists
         """
@@ -386,7 +450,7 @@ class Indexer(object):
                 # If that path already exists, but is not a directory, also fail
                 raise
 
-    def _test_wrap_apt_cache(self, wrapper):
+    def _test_wrap_apt_cache(self, wrapper) -> None:
         """
         Wrap the apt-cache in some proxy object.
 
@@ -399,7 +463,7 @@ class Indexer(object):
         # Wrap it
         self.apt_cache = wrapper(self.apt_cache)
 
-    def aptcache(self):
+    def aptcache(self) -> apt.Cache:
         if not self.apt_cache:
             #import warnings
             ## Yes, apt, thanks, I know, the api isn't stable, thank you so very much
@@ -407,15 +471,16 @@ class Indexer(object):
             #warnings.filterwarnings("ignore","apt API not stable yet")
             import apt
             import apt_pkg
+
             #warnings.resetwarnings()
 
             if self.quietapt:
                 class AptSilentProgress(apt.progress.text.OpProgress) :
-                    def __init__(self): 
+                    def __init__(self) -> None:
                         pass
-                    def done(self):
+                    def done(self) -> None:
                         pass
-                    def update(self,percent=None):
+                    def update(self, percent: float | None = None) -> None:
                         pass
                 aptprogress = AptSilentProgress()
             else:
@@ -425,7 +490,7 @@ class Indexer(object):
             self.apt_cache = apt.Cache(progress=aptprogress)
         return self.apt_cache
 
-    def lock(self):
+    def lock(self) -> bool:
         """
         Lock the session to prevent further updates.
 
@@ -446,7 +511,7 @@ class Indexer(object):
             else:
                 raise
 
-    def slave(self):
+    def slave(self) -> None:
         """
         Attach to a running indexer and report its progress.
 
@@ -456,7 +521,7 @@ class Indexer(object):
         childProgress = ClientProgress(self.progress)
         childProgress.loop()
 
-    def setupIndexing(self, force=False, system=True):
+    def setupIndexing(self, force: bool = False, system: bool = True) -> bool:
         """
         Setup indexing: read plugins, check timestamps...
 
@@ -521,10 +586,11 @@ class Indexer(object):
 
         return True
 
-    def get_document_from_apt(self, pkg):
+    def get_document_from_apt(self, pkg: apt.Package) -> xapian.Document:
         """
         Get a xapian.Document for the given apt package record
         """
+        assert self.plugins is not None
         document = xapian.Document()
         # The document data is the package name
         document.set_data(pkg.name)
@@ -538,10 +604,11 @@ class Indexer(object):
             addon.obj.index(document, pkg)
         return document
 
-    def get_document_from_deb822(self, pkg):
+    def get_document_from_deb822(self, pkg) -> xapian.Document:
         """
         Get a xapian.Document for the given deb822 package record
         """
+        assert self.plugins is not None
         document = xapian.Document()
 
         # The document data is the package name
@@ -556,7 +623,7 @@ class Indexer(object):
             addon.obj.indexDeb822(document, pkg)
         return document
 
-    def gen_documents_apt(self):
+    def gen_documents_apt(self) -> Iterator[xapian.Document]:
         """
         Generate Xapian documents from an apt cache
         """
@@ -566,39 +633,34 @@ class Indexer(object):
             if not pkg.candidate:
                 continue
             # Print progress
-            if idx % 200 == 0: self.progress.progress(100*idx/count)
+            if idx % 200 == 0:
+                self.progress.progress(100*idx/count)
             yield self.get_document_from_apt(pkg)
 
-    def gen_documents_deb822(self, fnames):
-        try:
-            from debian import deb822
-        except ImportError:
-            from debian_bundle import deb822
+    def gen_documents_deb822(self, fnames: Iterable[str]) -> Iterator[xapian.Document]:
+        from debian import deb822
         seen = set()
         for fname in fnames:
             infd = open(fname)
-            # In Python 3, you cannot tell() on a file object while
-            # iterating so you have to pop the hood to get at the
-            # underlying buffer.  In Python 2, the file object doesn't have
-            # this attribute, or this restriction.
-            try:
-                tellable = infd.buffer
-            except AttributeError:
-                tellable = infd
             # Get file size to compute progress
             total = os.fstat(infd.fileno())[6]
             for idx, pkg in enumerate(deb822.Deb822.iter_paragraphs(infd)):
                 name = pkg["Package"]
-                if name in seen: continue
+                if name in seen:
+                    continue
                 seen.add(name)
                 # Print approximate progress by checking the current read position
                 # against the file size
                 if total > 0 and idx % 200 == 0:
-                    cur = tellable.tell()
+                    cur = infd.buffer.tell()
                     self.progress.progress(100*cur/total)
                 yield self.get_document_from_deb822(pkg)
 
-    def compareCacheToDb(self, cache, db):
+    def compareCacheToDb(
+        self,
+        cache: apt.Cache,
+        db: xapian.WritableDatabase
+    ) -> tuple[dict[str, int], dict[str, int], dict[str, int]]:
         """
         Compare the apt cache to the database and return dicts
         of the form (pkgname, docid) for the following states:
@@ -607,16 +669,17 @@ class Indexer(object):
         outdated - a new version since the last update
         obsolete - no longer in the apt cache
         """
-        unchanged = {}
-        outdated = {}
-        obsolete = {}
+        unchanged: dict[str, int] = {}
+        outdated: dict[str, int] = {}
+        obsolete: dict[str, int] = {}
         self.progress.begin("Reading Xapian index")
         count = db.get_doccount()
         for (idx, m) in enumerate(db.postlist("")):
-            if idx % 5000 == 0: self.progress.progress(100*idx/count)
+            if idx % 5000 == 0:
+                self.progress.progress(100*idx/count)
             doc = db.get_document(m.docid)
             pkg = doc.get_data()
-            if bytes is not str and isinstance(pkg, bytes):
+            if isinstance(pkg, bytes):
                 # python3-apt requires strings.
                 pkg = pkg.decode('utf-8')
             # this will return '' if there is no value 0, which is fine because it
@@ -635,10 +698,11 @@ class Indexer(object):
         self.progress.end()
         return unchanged, outdated, obsolete
 
-    def updateIndex(self, pathname):
+    def updateIndex(self, pathname: str) -> None:
         """
         Update the index
         """
+        assert self.plugins is not None
         db = xapian.WritableDatabase(pathname, xapian.DB_CREATE_OR_OPEN)
         cache = self.aptcache()
         count = len(cache)
@@ -650,7 +714,8 @@ class Indexer(object):
                                                     len(obsolete)))
 
         self.progress.begin("Updating Xapian index")
-        for a in self.plugins: a.send_extra_info(db=db, aptcache=cache)
+        for a in self.plugins:
+            a.send_extra_info(db=db, aptcache=cache)
         for idx, pkg in enumerate(cache):
             if idx % 1000 == 0: self.progress.progress(100*idx/count)
             if not pkg.candidate:
@@ -674,7 +739,7 @@ class Indexer(object):
         db.flush()
         self.progress.end()
 
-    def incrementalUpdate(self):
+    def incrementalUpdate(self) -> None:
         if not os.path.exists(axi.XAPIANINDEX):
             self.progress.notice("No Xapian index built yet: falling back to full rebuild")
             return self.rebuild()
@@ -692,10 +757,11 @@ class Indexer(object):
         if self.ds_timestamp > 0:
             os.utime(axi.XAPIANDBSTAMP, (self.ds_timestamp, self.ds_timestamp))
 
-    def buildIndex(self, pathname, documents, addoninfo={}):
+    def buildIndex(self, pathname: str, documents, addoninfo={}) -> None:
         """
         Create a new Xapian index with the content provided by the plugins
         """
+        assert self.plugins is not None
         self.progress.begin("Rebuilding Xapian index")
 
         # Create a new Xapian index
@@ -703,19 +769,21 @@ class Indexer(object):
         # It seems to be faster without transactions, at the moment
         #db.begin_transaction(False)
 
-        for a in self.plugins: a.send_extra_info(db=db)
+        for a in self.plugins:
+            a.send_extra_info(db=db)
 
         # Add all generated documents to the index
         for doc in documents:
             db.add_document(doc)
 
         #db.commit_transaction();
-        for a in self.plugins: 
+        for a in self.plugins:
             a.finished()
         db.flush()
         self.progress.end()
 
-    def rebuild(self, pkgfiles=None):
+    def rebuild(self, pkgfiles: Iterable[str] | None = None) -> None:
+        assert self.plugins is not None
         # Create a new Xapian index with the content provided by the plugins
         # Xapian takes care of preventing concurrent updates and removing the old
         # database if it's left over by a previous crashed update
@@ -729,7 +797,8 @@ class Indexer(object):
         if pkgfiles:
             generator = self.gen_documents_deb822(pkgfiles)
         else:
-            for a in self.plugins: a.send_extra_info(aptcache=self.aptcache())
+            for a in self.plugins:
+                a.send_extra_info(aptcache=self.aptcache())
             generator = self.gen_documents_apt()
         self.buildIndex(dbdir, generator)
 
@@ -742,7 +811,7 @@ class Indexer(object):
         os.rename(axi.XAPIANINDEX + ".tmp", axi.XAPIANINDEX)
 
         # Remove all other index.* directories that are not the newly created one
-        def cleanoldcaches(dir):
+        def cleanoldcaches(dir: str) -> None:
             for file in os.listdir(dir):
                 if not file.startswith("index."): continue
                 # Don't delete what we just created
@@ -765,10 +834,12 @@ class Indexer(object):
         self.writePrefixes()
         self.writeDoc()
 
-    def writePrefixes(self, pathname=axi.XAPIANDBPREFIXES):
+    def writePrefixes(self, pathname: str = axi.XAPIANDBPREFIXES) -> None:
         """
         Write the prefix information on the given file
         """
+        assert self.plugins is not None
+
         self.progress.verbose("Writing prefix information to %s." % pathname)
         with open(pathname+".tmp", "w") as out:
             print(textwrap.dedent("""
@@ -785,7 +856,7 @@ class Indexer(object):
         """).lstrip(), file=out)
 
             # Aggregate and normalise prefix information from all the plugins
-            prefixes = dict()
+            prefixes: dict[str, dict[str, str]] = dict()
             for addon in self.plugins:
                 for p in addon.info.get("prefixes", []):
                     idx = p.get("idx", None)
@@ -809,7 +880,7 @@ class Indexer(object):
         # Atomic update of the documentation
         os.rename(pathname+".tmp", pathname)
 
-    def writeValues(self, pathname=axi.XAPIANDBVALUES):
+    def writeValues(self, pathname: str = axi.XAPIANDBVALUES) -> None:
         """
         Write the value information on the given file
         """
@@ -837,10 +908,11 @@ class Indexer(object):
         # Atomic update of the documentation
         os.rename(pathname+".tmp", pathname)
 
-    def writeDoc(self, pathname=axi.XAPIANDBDOC):
+    def writeDoc(self, pathname: str = axi.XAPIANDBDOC) -> None:
         """
         Write the documentation in the given file
         """
+        assert self.plugins is not None
         self.progress.verbose("Writing documentation to %s." % pathname)
         # Collect the documentation
         docinfo = []
